#For questions, please contact SDET project lead: Ruisheng.Diao@pnnl.gov
#August, 2017

import os, sys
import csv
from math import log10
PSSE_PATH = 'C:\\Program Files (x86)\\PTI\\PSSE33\\PSSBIN'
sys.path.append(PSSE_PATH)
os.environ['PATH'] += ';' + PSSE_PATH
#** Initialize PSSE modules
import psspy
import redirect
import random
import copy
#redirect.psse2py()
psspy.psseinit(1000000)

# change input file name here:
casenamefull = 'MMWG_2026SUM_2015Series_Final.sav'  #EI case
#casenamefull = '25hs1ap_wecc.sav'   # WECC case
#casenamefull = 'pfoutput_cleaned_PSSE33_new_7.raw' # PJM case

filetype_wrk = casenamefull[-4:] # last four characters
casename = casenamefull[:-4]

if filetype_wrk == '.sav' or filetype_wrk == '.SAV':
    psspy.case(casenamefull)
elif filetype_wrk == '.raw' or filetype_wrk == '.RAW':
    ierr = psspy.read(0, casenamefull)

logfile = 'SDET_zones_fragmentation_'+casename+'.log'
psspy.progress_output(2,logfile,[0,0])
logfile1 = 'SDET_zones_fragmentation_'+casename+'.log'
psspy.report_output(2,logfile1,[0,0])

_i = psspy.getdefaultint()
_f = psspy.getdefaultreal()
_s = psspy.getdefaultchar()

# ===================================== Area, Zone, Owner Info Anonymization, Qian Zhang =====================================

# User Defined Parameters
new_area_num = 1 # anonymized area number
new_area_name = 'AREA1' # anonymized area name
new_owner_num = 100 # anonymized owner number
new_owner_name = 'PNNL' # anonymized owner name

# Initialization
numofbus_zonechange_gen = 0 # to count the number of buses with zone num change for generator buses
numofbus_areachange_gen = 0 # to count the number of buses with area num change for generator buses
numofbus_zonechange_load = 0 # to count the number of buses with zone num change for load buses
numofbus_areachange_load = 0 # to count the number of buses with area num change for load buses

# Get old zone numbers and shuffle to create new zone numbers
ierr, iarray = psspy.azoneint(-1,2,'NUMBER')
izone_old = iarray[0] # existing zone number list
izone_tmp = copy.copy(izone_old) # copy the value of izone_old
random.seed(4) # To make sure the randomized sequence is the same each time
random.shuffle(izone_tmp)
izone_new = izone_tmp # new zone number list after random shuffle

# Get bus info
ierr, iarray = psspy.abusint(-1,2,'NUMBER')
ibus = iarray[0]
ierr, iarray = psspy.abusint(-1,2,'TYPE')
ibus_type = iarray[0]
ierr, iarray = psspy.abusint(-1,2,'AREA')
ibus_area = iarray[0]
ierr, iarray = psspy.abusint(-1,2,'ZONE')
ibus_zone = iarray[0]
businfo = zip(ibus,ibus_type,ibus_area,ibus_zone)

# Get load info
ierr, iarray = psspy.aloadint(-1,4,'NUMBER')
ibus_load = iarray[0]
ierr, carray = psspy.aloadchar(-1, 4, 'ID')
ibus_load_id = carray[0]
ierr, iarray = psspy.aloadint(-1,4,'AREA')
ibus_load_area = iarray[0]
ierr, iarray = psspy.aloadint(-1,4,'ZONE')
ibus_load_zone = iarray[0]
loadinfo = zip(ibus_load,ibus_load_id,ibus_load_area,ibus_load_zone)

# For each generator bus, change zone number to the new zone number and change area number to new_area_num.
for ibusinfo in businfo:
	zonenum_thisbus = ibusinfo[3]
	zoneindex_thisbus = izone_old.index(zonenum_thisbus)
	ierr = psspy.bus_data_3(ibusinfo[0],intgar3 = izone_new[zoneindex_thisbus]) # zone number change
	if ierr == 0:
		numofbus_zonechange_gen += 1

	ierr = psspy.bus_data_3(ibusinfo[0],intgar2 = new_area_num) # area number change
	if ierr <=0: # ignore warning messages of resetting area numbers for the area swing buses
		numofbus_areachange_gen += 1

# For each load bus, change zone number to the new zone number and change area number to new_area_num.
for iloadbusinfo in loadinfo:
	zonenum_thisbus = iloadbusinfo[3]
	zoneindex_thisbus = izone_old.index(zonenum_thisbus)
	ierr = psspy.load_chng_4(iloadbusinfo[0],iloadbusinfo[1],intgar3 = izone_new[zoneindex_thisbus]) # zone number change
	if ierr == 0:
		numofbus_zonechange_load += 1

	ierr = psspy.load_chng_4(iloadbusinfo[0],iloadbusinfo[1],intgar2 = new_area_num) # area number change
	if ierr <=0: # ignore warning messages of resetting area numbers for the area swing buses
		numofbus_areachange_load += 1

# Change zone names in the zones table for the new zone numbers
ierr, iarray = psspy.azoneint(-1,1,'NUMBER')
izone_crt = iarray[0]
izonename_crt = []
for izonenum in izone_crt:
	izonename_crt.append("Z"+str(izonenum))
for izonenum in izone_crt:
	izone_crt_index = izone_crt.index(izonenum)
	ierr = psspy.zone_data(izonenum,izonename_crt[izone_crt_index])

# Change all area names to new_area_name
ierr, iarray = psspy.aareaint(-1,1,'NUMBER') # note the second number is 1
iarea_crt = iarray[0] # should only contain area 1 to this point
for iareanum in iarea_crt:
	ierr = psspy.area_data(iareanum,arname = new_area_name)

# Delete areas except for new_area_num with no equipment assigned to them
ierr, iarray = psspy.aareaint(-1,2,'NUMBER') # note the second number is 2
iarea = iarray[0]
if new_area_num in iarea:
	index_new_area_num = iarea.index(new_area_num)
	del iarea[index_new_area_num]
ierr = psspy.asysinit(1)
ierr = psspy.asys(1,len(iarea),iarea)
ierr = psspy.purgarea(1,0)

# Change the area new_area_num swing bus to system swing bus
for ibusinfo in businfo:
	if ibusinfo[1] == 3:
		sys_swing_bus = ibusinfo[0]
		break
ierr = psspy.area_data(new_area_num,intgar1 = sys_swing_bus)

# Set desired interchange of area 1 to 0 and the desired MW interchange tolerance to 10.0
ierr = psspy.area_data(1,realar1 = 0.0,realar2 = 10.0)

# Owner renumbering for all components
ierr = psspy.ownm_3(0,1,[1,1,1,1,1,1,1,1,1],new_owner_num)

# Change owner names to new_owner_name
ierr, iarray = psspy.aownerint(-1,2,'NUMBER') # note the second number is 2
iowner_crt = iarray[0] # should only contain new_owner_num to this point
for iownernum in iowner_crt:
	ierr = psspy.owner_data(iownernum,owname = new_owner_name)

# Print messages and save a new raw file with anonymized zone and area info
if numofbus_zonechange_gen == len(ibus) and numofbus_zonechange_load == len(ibus_load):
	print "********** Zone info of all buses have been successfullly anonymized. *************"
	print "Original zone numbers are:"
	print izone_old
	print "New zone numbers are:"
	print izone_new
else:
	print "********** Some error occurred when changing some bus zone numbers. ***********"

if numofbus_areachange_gen == len(ibus) and numofbus_areachange_load == len(ibus_load):
	print "********** area info of all buses have been successfullly anonymized. *************"
	print "All area numbers are set to:",new_area_num
	print "All area names are set to:",new_area_name
else:
	print "********** Some error occurred when changing some bus area numbers. ***********"

if len(iowner_crt)==1:
	print "********** area owner of all buses have been successfullly anonymized. *************"
	print "All owner numbers are set to:",new_owner_num
	print "All owner names are set to:",new_owner_name
else:
	print "********** Not all owner info is anonymized. Please check. ***********"

banony = 0
if numofbus_zonechange_gen == len(ibus) and numofbus_areachange_gen == len(ibus) and numofbus_zonechange_load == len(ibus_load) and numofbus_areachange_load == len(ibus_load) and len(iowner_crt)==1:
	banony = 1
	psspy.rawd_2(0,1,[1,1,1,0,0,0,0],0,casename+'_anony'+'.raw')
	print "New raw file saved with anonymized area,zone and owner info."

print "Next, creating fragments, please wait..."
# ======================================================================================================================
#
ierr,iarray = psspy.amachint(-1,4,'NUMBER')
igenbus = iarray[0]
ierr,sarray = psspy.amachchar(-1,4,'ID')
genid = sarray[0]
ierr,iarray = psspy.aloadint(-1,4,'NUMBER')
iloadbus = iarray[0]
ierr,sarray = psspy.aloadchar(-1,4,'ID')
loadid = sarray[0]

# Checking for generators and loads with ID '99' and changing them
for ii in range(0,len(igenbus)):
	if genid[ii] == '99':
		for jj in range(98,1,-1):
			ierr,temp = psspy.macdat(igenbus[ii],str(jj),'P')
			if ierr == 2:
				break
		psspy.mbidmac(igenbus[ii],'99',str(jj))

for ii in range(0,len(iloadbus)):
	if loadid[ii] == '99':
		for jj in range(98,1,-1):
			ierr,temp = psspy.loddt1(iloadbus[ii],str(jj),'MVA','ACT')
			if ierr == 2:
				break
		psspy.mbidload(iloadbus[ii],'99',str(jj))

casename1 = casename+'1'
psspy.save(casename1+'.sav')


ierr, iarray = psspy.azoneint(-1,1,'NUMBER')
izone = iarray[0]
ierr, iarray = psspy.abusint(-1,1,'NUMBER')
ibus = iarray[0]
ierr, iarray = psspy.abusint(-1,1,'ZONE')
ibuszone = iarray[0]
ierr, iarray = psspy.abrnint(-1,1,3,2,1,'FROMNUMBER')
linefbus = iarray[0]
ierr, iarray = psspy.abrnint(-1,1,3,2,1,'TONUMBER')
linetbus = iarray[0]
ierr, sarray = psspy.abrnchar(-1,1,3,2,1,'ID')
lineid = sarray[0]
ierr, iarray = psspy.atrnint(-1,1,3,2,1,'FROMNUMBER')
xfrfbus = iarray[0]
ierr, iarray = psspy.atrnint(-1,1,3,2,1,'TONUMBER')
xfrtbus = iarray[0]
ierr, sarray = psspy.atrnchar(-1,1,3,2,1,'ID')
xfrid = sarray[0]
ierr, iarray = psspy.atr3int(-1,1,3,2,1,'WIND1NUMBER')
xfr1bus = iarray[0]
ierr, iarray = psspy.atr3int(-1,1,3,2,1,'WIND2NUMBER')
xfr2bus = iarray[0]
ierr, iarray = psspy.atr3int(-1,1,3,2,1,'WIND3NUMBER')
xfr3bus = iarray[0]
ierr, sarray = psspy.atr3char(-1,1,3,2,1,'ID')
xfr3id = sarray[0]

newbusnum = max(ibus)+100

for izonenum in izone:
	psspy.fnsl([0,0,0,1,1,0,99,0])
	for ii in range(0,len(linefbus)):
		ierr, rval = psspy.brndat(linefbus[ii],linetbus[ii],lineid[ii],'RATEA')
		if ierr == 0:
			ierr, zonefbus = psspy.busint(linefbus[ii],'ZONE')
			ierr, zonetbus = psspy.busint(linetbus[ii],'ZONE')
			if zonefbus == izonenum and zonetbus != izonenum:
				breakbranch = 1;
			elif zonetbus == izonenum and zonefbus != izonenum:
				breakbranch = 1;
			else:
				breakbranch = 0;
			if breakbranch == 1:
				ierr, kv1 = psspy.busdat(linefbus[ii],'BASE')
				ierr, kv2 = psspy.busdat(linetbus[ii],'BASE')
				if kv1 >= kv2:
					psspy.splt(linefbus[ii],newbusnum,'B'+str(newbusnum), kv1)
					psspy.movebrn(linetbus[ii],linefbus[ii],lineid[ii],newbusnum,lineid[ii])
					psspy.bus_chng_3(newbusnum,[_i,_i,zonetbus,_i],[_f,_f,_f,_f,_f,_f,_f],_s)
					newbusnum = newbusnum + 1
				else:
					psspy.splt(linetbus[ii],newbusnum,'B'+str(newbusnum), kv2)
					psspy.movebrn(linefbus[ii],linetbus[ii],lineid[ii],newbusnum,lineid[ii])
					psspy.bus_chng_3(newbusnum,[_i,_i,zonefbus,_i],[_f,_f,_f,_f,_f,_f,_f],_s)
					newbusnum = newbusnum + 1
	for ii in range(0,len(xfrfbus)):
		ierr, rval = psspy.xfrdat(xfrfbus[ii],xfrtbus[ii],xfrid[ii],'RATIO')
		if ierr == 0:
			ierr, zonefbus = psspy.busint(xfrfbus[ii],'ZONE')
			ierr, zonetbus = psspy.busint(xfrtbus[ii],'ZONE')
			if zonefbus == izonenum and zonetbus != izonenum:
				breakbranch = 1;
			elif zonetbus == izonenum and zonefbus != izonenum:
				breakbranch = 1;
			else:
				breakbranch = 0;
			if breakbranch == 1:
				ierr, kv1 = psspy.busdat(xfrfbus[ii],'BASE')
				ierr, kv2 = psspy.busdat(xfrtbus[ii],'BASE')
				if kv1 >= kv2:
					psspy.splt(xfrfbus[ii],newbusnum,'B'+str(newbusnum), kv1)
					psspy.movebrn(xfrtbus[ii],xfrfbus[ii],xfrid[ii],newbusnum,xfrid[ii])
					psspy.bus_chng_3(newbusnum,[_i,_i,zonetbus,_i],[_f,_f,_f,_f,_f,_f,_f],_s)
					newbusnum = newbusnum + 1
				else:
					psspy.splt(xfrtbus[ii],newbusnum,'B'+str(newbusnum), kv2)
					psspy.movebrn(xfrfbus[ii],xfrtbus[ii],xfrid[ii],newbusnum,xfrid[ii])
					psspy.bus_chng_3(newbusnum,[_i,_i,zonefbus,_i],[_f,_f,_f,_f,_f,_f,_f],_s)
					newbusnum = newbusnum + 1

psspy.save(casename1+'1.sav')

ierr, iarray = psspy.azoneint(-1,1,'NUMBER')
izone = iarray[0]
ierr, iarray = psspy.abusint(-1,1,'NUMBER')
ibus = iarray[0]
ierr, iarray = psspy.abusint(-1,1,'ZONE')
ibuszone = iarray[0]
ierr, iarray = psspy.abrnint(-1,1,3,2,1,'FROMNUMBER')
linefbus = iarray[0]
ierr, iarray = psspy.abrnint(-1,1,3,2,1,'TONUMBER')
linetbus = iarray[0]
ierr, sarray = psspy.abrnchar(-1,1,3,2,1,'ID')
lineid = sarray[0]
ierr, iarray = psspy.atrnint(-1,1,3,2,1,'FROMNUMBER')
xfrfbus = iarray[0]
ierr, iarray = psspy.atrnint(-1,1,3,2,1,'TONUMBER')
xfrtbus = iarray[0]
ierr, sarray = psspy.atrnchar(-1,1,3,2,1,'ID')
xfrid = sarray[0]
ierr, iarray = psspy.atr3int(-1,1,3,2,1,'WIND1NUMBER')
xfr1bus = iarray[0]
ierr, iarray = psspy.atr3int(-1,1,3,2,1,'WIND2NUMBER')
xfr2bus = iarray[0]
ierr, iarray = psspy.atr3int(-1,1,3,2,1,'WIND3NUMBER')
xfr3bus = iarray[0]
ierr, sarray = psspy.atr3char(-1,1,3,2,1,'ID')
xfr3id = sarray[0]


for izonenum in izone:
	print' create fragment for zone number: '+str(izonenum)
	psspy.case(casename1+'1.sav')
	psspy.fnsl([0,0,0,1,1,0,99,0])
	for ii in range(0,len(linefbus)):
		ierr, zonefbus = psspy.busint(linefbus[ii],'ZONE')
		ierr, zonetbus = psspy.busint(linetbus[ii],'ZONE')
		if zonefbus == izonenum and zonetbus != izonenum:
			breakbranch = 1;
		elif zonetbus == izonenum and zonefbus != izonenum:
			breakbranch = 1;
		else:
			breakbranch = 0;
		if breakbranch == 1:
			psspy.branch_chng(linefbus[ii],linetbus[ii],lineid[ii],[0,_i,_i,_i,_i,_i],[_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f])
	for ii in range(0,len(xfrfbus)):
		ierr, zonefbus = psspy.busint(xfrfbus[ii],'ZONE')
		ierr, zonetbus = psspy.busint(xfrtbus[ii],'ZONE')
		if zonefbus == izonenum and zonetbus != izonenum:
			breakbranch = 1;
		elif zonetbus == izonenum and zonefbus != izonenum:
			breakbranch = 1;
		else:
			breakbranch = 0;
		if breakbranch == 1:
			aa = 1
			psspy.two_winding_chng_4(xfrfbus[ii],xfrtbus[ii],xfrid[ii],[0,_i,_i,_i,_i,_i,_i,_i,_i,_i,_i,_i,_i,_i,_i],[_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f],[_s,""])
	for ii in range(0,len(xfr1bus)):
		ierr, zone1bus = psspy.busint(xfr1bus[ii],'ZONE')
		ierr, zone2bus = psspy.busint(xfr2bus[ii],'ZONE')
		ierr, zone3bus = psspy.busint(xfr3bus[ii],'ZONE')
		if zone1bus == izonenum:
			breakbranch1 = 1;
		else:
			breakbranch1 = 0;
		if zone2bus == izonenum:
			breakbranch2 = 1;
		else:
			breakbranch2 = 0;
		if zone3bus == izonenum:
			breakbranch3 = 1;
		else:
			breakbranch3 = 0;
		totbreakbranch = breakbranch1 + breakbranch2 + breakbranch3
		if totbreakbranch == 1 or totbreakbranch == 2:
			psspy.three_wnd_imped_chng_3(xfr1bus[ii],xfr2bus[ii],xfr3bus[ii],xfr3id[ii],[_i,_i,_i,_i,_i,_i,_i,0,_i,_i,_i,_i],[_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f,_f],[_s,""])
	ierr = psspy.bgen(0,1,0)
	psspy.bsys(1,0,[0, 999],0,[],0,[],0,[],1,[izonenum])
	psspy.rawd_2(1,0,[0,1,1,0,0,0,0],0,'test.raw') # Qian Zhang: changed the first number in the third element to 0 to eliminate type 4 buses
	psspy.read(0,'test.raw')

	# if len(genpmax)>0:
	# 	slackbus = igenbus[genpmax.index(max(genpmax))]
	# 	psspy.bus_chng_3(slackbus,[3,_i,_i,_i],[_f,_f,_f,_f,_f,_f,_f],_s)
		#psspy.island()

	# ============ Find the bus that has largest generator (not '99') and using it as slackbus, Qian Zhang==============
	ierr,iarray = psspy.amachint(-1,1,'NUMBER')
	igenbus = iarray[0]
	ierr,rarray = psspy.amachreal(-1,1,'PMAX')
	genpmax = rarray[0]
	ierr,rarray = psspy.amachreal(-1,1,'PGEN')
	genpgen = rarray[0]
	
	
	if not igenbus:
		continue

	geninfo = zip(igenbus,genpmax,genpgen)
	gen_tmp_sorted = sorted(geninfo, key=lambda item:item[2], reverse=True) # sort based on the pgen of the slack bus

	bno99gen = 0
	for ii in range(0,len(gen_tmp_sorted)):
		# Set one bus as a subsystem "1" for each iteration
		temp_bus = gen_tmp_sorted[ii][0]
		psspy.bsys(1,0,[0.,1000.],0,[],1,[temp_bus],0,[],0,[])
		# Check if this bus (subsystem) has 99 generator
		ierr, iarray_genid = psspy.amachchar(1,1,'ID') # get gen id on this bus (subsystem "1")
		allgenid_tmp = iarray_genid[0]
		if '99' not in allgenid_tmp:
			slackbus = temp_bus
			print ('slack bus choose for zone: ' + str(izonenum) + 'bus no: ' + str (slackbus))
			psspy.bus_chng_3(slackbus,[3,_i,_i,_i],[_f,_f,_f,_f,_f,_f,_f],_s)
			bno99gen = 1
			break
	
	if bno99gen==0:
		continue

	# ==================================================================================================================

	# ====================== Relax Q limits for generators with ID '99', Qian Zhang ====================================
	ierr,iarray = psspy.amachint(-1,1,'NUMBER')
	igenbus = iarray[0]
	ierr, carray = psspy.amachchar(-1, 1, 'ID')
	igenid = carray[0]
	geninfo_4Qrlx = zip(igenbus,igenid)

	for item in geninfo_4Qrlx:
		if item[1] == '99':
			ierr = psspy.machine_data_2(item[0], item[1], realar3 = 600, realar4 = -600)
	# ==================================================================================================================
	
	psspy.island()

	psspy.fnsl([0,0,0,1,0,0,99,0])

	ival = psspy.solved()
	if ival == 0:
		ierr,totbusnum = psspy.abuscount(-1,2)
		ierr,totgennum = psspy.amachcount(-1,4)
		ierr,totpload = psspy.azonereal(-1,2,'PLOAD')
		ierr,totpgen = psspy.azonereal(-1,2,'PGEN')
		ierr,totbrnnum = psspy.abrncount(-1,1,3,2,1)
		ierr,tot2wdgnum = psspy.atrncount(-1,1,3,2,1)
		if totbusnum > 5:
			psspy.rawd_2(0,1,[0,1,1,0,0,0,0],0,casename+'_Zone'+str(izonenum)+'_Busnum'+str(totbusnum)+'_Gennum'+str(totgennum)+'_load'+str(int(totpload[0][0]))+'MW_Gen'+str(int(totpgen[0][0]))+'MW_brn'+str(totbrnnum)+'_Trn'+str(tot2wdgnum)+'.raw')
	#print ' bgen error '+str(ierr)

#remove the temp files
os.remove(casename+r'1.sav')
os.remove(casename+r'11.sav')
os.remove(r'test.raw')
	
if banony == 1:
	os.remove(casename+r'_anony.raw')
	
print ('finished fragmentation!!')
